
/*
	dso.c - suppot functions for dynamic shared objects (modules)
*/





/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include <dlfcn.h>

#include "dlgbox.h"
#include "interface.h"
#include "gtkhelpers.h"
#include "import.h"
#include "functions.h"


/*******************************************************

	Load all DSO's found in ~/dso

********************************************************/

/* TODO: Specify the DSO path through the preferences */


int load_dso_singlepath (char* dsostr)
{
void* handle;
void (*initfn)();
int dsocnt;
struct stat statresult;		/* to check if a file is a directory */
DIR* directory;
struct dirent* entry;
gchar dir[1024]; gchar filemask[256]; gchar resfile[2048];
int i,j,k,err;
char dsopath2[512];


	dsocnt=0; strcpy (dsopath2, dsostr);
	if (strlen(dsostr)==0) return 0;	/* return if no path specified */

	/* Prepare the dso path and file names */

	stat (dsopath2, &statresult);
	if (S_ISDIR (statresult.st_mode))
	{
		if (dsopath2[strlen(dsopath2)-1]!='/')
			strcat (dsopath2, "/");	/* all in this directory */
		strcat (dsopath2, "*.dso");	/* all so's in this directory */
	}
	dp (2, "DSO(s) selected: %s\n", dsopath2);

	/* Now split dsopath into directory and file spec */

	i = strlen(dsopath2)-1;
	while ((i>0) && (dsopath2[i]!='/')) i--;
	k=0;
	for (j=0; j<i; j++) dir[k++]=dsopath2[j];
	dir[k]=0;	/* Null-terminate */
	k=0;
	for (j=i+1; j<strlen(dsopath2); j++) filemask[k++]=dsopath2[j];
	filemask[k]=0;
	dp (2, "Split: %s --- %s\n",dir,filemask);
	
	/* Scan the directory for matching files */
	
	directory = opendir (dir);
	
	if (!directory) 
	{
		dp (1, "Directory open failed %s, errcode %d\n",dsostr,errno);
		return 0;
	}
	
	do
	{
		entry = readdir (directory);
		if (entry) 
		{
			strcpy (resfile, dir); strcat (resfile, "/");
			strcat (resfile, entry->d_name);
			err=stat (resfile, &statresult);
			if ( (!S_ISDIR (statresult.st_mode))	/* Skip directories */
				&& (!err)							/* error STATing */
				&& (shmatch (filemask, entry->d_name)))
			{
				dp (2, "Loading DSO: %s\n", resfile);

				handle = dlopen (resfile, RTLD_NOW); 		/* was RTLD_LAZY */
				if (!handle)
				{
					dp (0,"DSO load failed. Reason: %s\n", dlerror());
				}
				else
				{
					initfn = dlsym (handle, "initmodule"); 
					(*initfn)();
					dsocnt++;
					dp (1, "DSO load successful\n");
				}

			}
		}
	}
	while (entry);
	dp (1, "%d DSOs loaded from %s\n",dsocnt, dsostr);
		
	return dsocnt;
}


/* Separate paths by their ; and call each location individually */

int load_dso ()
{
int total_dso,i, j, k;
char dsostr[512];


	total_dso=0;
	
	/* Default location */
	
	strcpy (dsostr, "/usr/local/cimage/dso");
	total_dso += load_dso_singlepath (dsostr);

	/* and the location(s) in the preferences */

	i=0;
	while (i<strlen(dsopath))
	{
		j=i;
		while ((j<strlen(dsopath)) && (dsopath[j]!=';')) j++;
		k=0;
		while (i<j) dsostr[k++]=dsopath[i++];
		dsostr[k]=0;
		total_dso += load_dso_singlepath (dsostr);
		i+=2;
	}
	return total_dso;
}



/*******************************************************

	This function is called by the DSO initializer
	to register a new menu option in the main menu
	(menu item "Special" only)

********************************************************/


/* Register a menu item. The menu item appears as "menuitemname" under
   "Special". The callback function gets called if this item is activated */

void dso_menu_register (gchar* menuitemname, void* callback)
{
GtkWidget* menuitem;

	menuitem = Menuitem (window1, mspecial_menu, NULL, menuitemname, menuitemname);
	gtk_signal_connect (GTK_OBJECT (menuitem), "activate", GTK_SIGNAL_FUNC (callback), NULL);


}


/* Register a menu item that is supposed to have a submenu. The item appears
   under "Special". There is no callback function. This function returns
   a pointer to the menu which is used by dso_submenu_register */

GtkWidget* dso_menu_register_with_submenu (gchar* menuitemname)
{
GtkWidget *menuitem, *m_menu;
GtkAccelGroup *menu_accels;

	menuitem = Menuitem (window1, mspecial_menu, NULL, menuitemname, menuitemname);
	m_menu = Makesubmenu (window1, menuitem, "m_menu");
  	///menu_accels = gtk_menu_ensure_uline_accel_group (GTK_MENU (m_menu));
 	return m_menu;

}

/* Register a submenu. Works like dso_menu_register, but wel need to pass
   a pointer to the parent menu item */

void dso_submenu_register (GtkWidget* toplevel, gchar* menuitemname, void* callback)
{
GtkWidget* menuitem;

 	menuitem = Menuitem (window1, toplevel, NULL, menuitemname, menuitemname);
	gtk_signal_connect (GTK_OBJECT (menuitem), "activate", GTK_SIGNAL_FUNC (callback), NULL);

}

