
/*
	magnifier.c - Magnifier window and Thumbnail window
*/


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


#include "callbacks.h"
#include "defs.h"
#include "application.h"
#include "magnifier.h"
#include "functions.h"
#include "imgbuf.h"


#define MAGSIZE 		33
#define MAGWINSIZE 		264
#define MAGFAC			8



image_type mag_img;


/*==============================================================*/

/* Depending on the type of image, create an RGB imlib image
   - the magnifier option ALWAYS expands contrast to fill the full 
     RGB range, irrespective of autoscale setting
     
     Note that xs, ys define the top left corner.
*/



void mag_fill8x8 (rgbtriplet *buf, rgbtriplet color, int x, int y, int stride)
{
int x0, y0, x1, y1;
int idx;


	x0 = x*MAGFAC; y0=y*MAGFAC;
	for (y1=y0; y1<y0+MAGFAC; y1++)
	{
		idx = x0+stride*y1;
		for (x1=0; x1<MAGFAC; x1++)
			buf[idx++] = color;
	}
}



int mag_get_image_from_data (image_type* srcimg, 
								int xs, int ys, int zs,int xsize0, int ysize0)
{
rgbtriplet* temp;		/* This is really a RGB structure */
rgbtriplet *p;
rgbtriplet color;
long idx;
int x,y,xsize,ysize;
int h,w,stride;


	if (!srcimg->pixbuf) return -1;

	xsize = xsize0*MAGFAC; ysize = ysize0*MAGFAC;

	temp= (rgbtriplet*) calloc ((xsize+1)*(ysize+1), sizeof (rgbtriplet));
	if (!temp) fatal_memerr();

	/* Get information from the main image's pixbuf */

	w = gdk_pixbuf_get_width(srcimg->pixbuf);
	h = gdk_pixbuf_get_height(srcimg->pixbuf);
	p = (rgbtriplet*)gdk_pixbuf_get_pixels (srcimg->pixbuf);
	dp (2, "Magnifier: Re-creating data from (%d,%d) to (%d,%d)\n", xs,ys,xs+xsize,ys+ysize);

	/* We now need to copy the RGB info pixel by pixel. Note that this function MAY be called
		with negative start coordinates or coordinates beyond the image window */

	for (y = ys; y<ys+ysize0; y++)
		for (x = xs; x<xs+xsize0; x++)
		{
			if ((x>=0) && (y>=0) && (x<w) && (y<h))
			{
				color = p[x+w*y];
				mag_fill8x8 (temp, color, x-xs, y-ys, xsize);
			}
			else
			{
				if ((x&1) ^ (y&1))	/* gray checkerboard pattern */
				{
					color.red=color.green=color.blue=150; 
				}
				else
				{
					color.red=color.green=color.blue=98; 
				}
				mag_fill8x8 (temp, color, x-xs, y-ys, xsize);
			}
		}

	/* Create an intermediate, unscaled pixbuf */

	mag_img.data = temp;
	mag_img.xmax = xsize; mag_img.ymax = ysize;
	mag_img.pixbuf = gdk_pixbuf_new_from_data ((unsigned char*)temp, GDK_COLORSPACE_RGB, FALSE, 8, 
														xsize, ysize, 3*xsize 		/* rowstride */,
														pixbuf_destroy_handler		/* destroy callback */,
														mag_img.pixbuf 				/* destroy callback user_data */);

	/* Create the final, scaled-up pixbuf. The original area is 33x33 pixels (fixed)
			and the magnified area is 8x magnified, that is, 264x264.
			Note that we NEED to use nearest-neighbor interpolation to get the pixelated effect */

	mag_img.width = xsize; mag_img.height = ysize;
	return 0;

}



/* TODO: This is still buggy. In some cases, the backcalculation of the main image coordinates
		is one pixel off. */


gboolean on_magnify_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
int x,y,ax,ay,ix,iy;
GdkModifierType state;
double magsize, midpointx, midpointy, boxsizex, boxsizey;

	gdk_window_get_pointer (widget->window, &x, &y, &state);

	/* Divide x,y by MAGFAC and shift origin to get original pixel offsets */

	ix = x/MAGFAC-MAGSIZE/2; iy = y/MAGFAC-MAGSIZE/2;

	/* We know that the origin (top left) of the magnifier coincides with mag_lastx,mag_lasty,
		so we need to add these to ix,iy to get main window's coordinates */

	dp (1, "Magnifier: Click at (%d %d)  size (%d %d)  scaled (%d %d)  final (%d %d)\n",
			x,y, mag_img.drawable->allocation.width,mag_img.drawable->allocation.height, ix,iy,
			mag_lastx+ix,mag_lasty+iy);
	ix += mag_lastx; iy += mag_lasty;
	if ((ix>=0) && (iy>=0) && (ix<mainimg.xmax) && (iy<mainimg.ymax))
		pixelpeek (ix, iy);

  	return FALSE;
}



gboolean on_magnify_expose_event (GtkWidget *widget, GdkEventExpose  *event, gpointer user_data)
{
	/* Pretty broken .How can we zoom? */

	if(event->count > 0) 			/* Draw only last expose */
	{
		return TRUE;
	}
	dp (2, "Magnifier expose. Area is %d, %d, %d, %d\n", event->area.x, event->area.y,
    			event->area.width, event->area.height);

	gdk_pixbuf_render_to_drawable (mag_img.pixbuf, mag_img.drawable->window, 
									mag_img.drawable->style->fg_gc[GTK_STATE_NORMAL], 
									event->area.x,event->area.y,				/* src_x and src_y */
									event->area.x,event->area.y,				/* dest_x and dest_y */
									event->area.width,event->area.height,		/* width & height */
									GDK_RGB_DITHER_NORMAL, 0, 0);

  	return FALSE;
}




gboolean on_magnify_configure_event (GtkWidget *widget, GdkEventButton *event, gpointer  user_data)
{
int x,y;

	if (!mag_img.data) return FALSE;

	x=mag_lastx-MAGSIZE/2; y=mag_lasty-MAGSIZE/2;

	mag_get_image_from_data (&mainimg, x,y,cur_slice, MAGSIZE,MAGSIZE);
	gtk_widget_set_size_request (mag_img.parent, MAGWINSIZE, MAGWINSIZE);

	gdk_flush ();

	return FALSE;
}



void create_magnifier (void)
{
GtkWidget *m_window;
GtkWidget *magnify;

	m_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (m_window), "m_window", m_window);
	gtk_window_set_title (GTK_WINDOW (m_window), "Magnifying glass");
	
	magnify = gtk_drawing_area_new ();
	gtk_widget_ref (magnify);
	gtk_object_set_data_full (GTK_OBJECT (m_window), "magnify", magnify,
								(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (magnify);
	gtk_container_add (GTK_CONTAINER (m_window), magnify);
	gtk_widget_set_events (magnify, GDK_BUTTON_PRESS_MASK);

	gtk_signal_connect (GTK_OBJECT (m_window), "destroy",
						GTK_SIGNAL_FUNC (destroy_magnifier),NULL);
	gtk_signal_connect (GTK_OBJECT (magnify), "button_press_event",
						GTK_SIGNAL_FUNC (on_magnify_button_press_event),NULL);
						
	gtk_signal_connect (GTK_OBJECT (magnify), "configure_event",
						GTK_SIGNAL_FUNC (on_magnify_configure_event),NULL);
	gtk_signal_connect (GTK_OBJECT (magnify), "expose_event",
						GTK_SIGNAL_FUNC (on_magnify_expose_event),NULL);
	
	gtk_widget_set_size_request (m_window, MAGWINSIZE, MAGWINSIZE);
	gtk_window_set_resizable (GTK_WINDOW(m_window), FALSE);		/* No user resizing allowed */
	gtk_widget_show (m_window);
	have_magnifier = 1; 

	gdk_window_set_cursor (magnify->window, crosshair_cursor);
	mag_img.parent = m_window;
	mag_img.drawable = magnify;

}


void destroy_magnifier ()
{
	have_magnifier = 0;
	freebuf (&mag_img);
	// if (mag_img.pixbuf) g_object_unref (G_OBJECT(mag_img.pixbuf));
	mag_img.pixbuf = NULL;

	gtk_widget_destroy ( GTK_WIDGET ( mag_img.parent ) );

}


/* Broken. This works only on mainimg (see configure event) */

void display_magnified (image_type* img, int x, int y, int z)
{
int w,h;


	if (!img->data) return;

	mag_lastx = x; mag_lasty = y;
	mag_get_image_from_data (img, mag_lastx-MAGSIZE/2,mag_lasty-MAGSIZE/2,z, MAGSIZE,MAGSIZE);
	if (!have_magnifier) 
		create_magnifier();
	else
	{
		w = mag_img.width;
		h = mag_img.height;
		dp (2, "Magnifier: Scheduled expose event\n");
		gtk_widget_queue_draw_area (mag_img.drawable, 0, 0, w, h);
	}


}






/*   #####   #   #   #   #   #   #   ####    #   #     #      ###    #       */
/*     #     #   #   #   #   ## ##   #   #   #   #    # #      #     #       */
/*     #     #   #   #   #   # # #   #   #   ##  #   #   #     #     #       */
/*     #     #####   #   #   # # #   ####    # # #   #   #     #     #       */
/*     #     #   #   #   #   #   #   #   #   #  ##   #####     #     #       */
/*     #     #   #   #   #   #   #   #   #   #   #   #   #     #     #       */
/*     #     #   #    ###    #   #   ####    #   #   #   #    ###    #####   */





int thumbnail_get_image_from_data (image_type* srcimg)
{
rgbtriplet* temp;		/* This is really a RGB structure */
rgbtriplet *p;
rgbtriplet color;
long idx;
int x,y,x1,y1,xsize,ysize;
int h,w,stride;
float xf, yf, sc, delta, min, max;


	if (!srcimg->data) return -1;
	imgminmax (srcimg, &min, &max);

	xsize = TN_SIZE, ysize = TN_SIZE;		/* Keep this flexible for a later adjustment of the aspect ratio */

	temp= (rgbtriplet*) calloc ((xsize+1)*(ysize+1), sizeof (rgbtriplet));
	if (!temp) fatal_memerr();

	/* Get information from the main image's pixbuf */

	w = gdk_pixbuf_get_width(srcimg->pixbuf);
	h = gdk_pixbuf_get_height(srcimg->pixbuf);
	p = (rgbtriplet*)gdk_pixbuf_get_pixels (srcimg->pixbuf);

	/* Get the larger dimension, the shrink ratio, and the step sizes */

	delta = (float)(MAX(w,h))/(float)TN_SIZE;
	dp (2, "Thumbnail: Re-creating from data, scaled down by %.0f\n", delta);

	/* Default fill is the checkerboard pattern */

	for (y=0; y<ysize; y++)
		for (x=0; x<xsize; x++)
		{
			if (( (x>>3) &1) ^ ( (y>>3) &1))	/* gray checkerboard pattern */
			{
				color.red=color.green=color.blue=150; 
			}
			else
			{
				color.red=color.green=color.blue=98; 
			}
			temp[x+xsize*y]=color;
		}


	/* We now need to copy the RGB info, scaling down to thumbnail size */

	y=0;
	for (yf = 0.0; yf < h; yf+=delta)
	{
		x=0;
		for (xf = 0.0; xf < w; xf+=delta)
		{

			/* Distinguish between RGB and non-RGB */

			if (mainimg.imgtype==B_RGB)
			{
				if ((xf>=0) && (yf>=0) && (xf<=srcimg->xmax-1) && (yf<=srcimg->ymax-1))
				{
					color = rgbireadbuf (&mainimg, xf,yf,cur_slice, INTP_NN);
					temp[x+xsize*y] = color;
				}
				
			}
			else
			{
				if ((xf>=0) && (yf>=0) && (xf<=srcimg->xmax-1) && (yf<=srcimg->ymax-1))
				{
					color.red = (int)(255.0*(ireadbuf (&mainimg, xf,yf,cur_slice, INTP_NN)-min)/(max-min));
					color.green = color.red; color.blue = color.red;
					temp[x+xsize*y] = color;
				}
			}
			x++;
		}
		y++;
	}

	/* Create the final pixbuf for the thumbnail image */

	tn_img.data = NULL;			/* Never has any real image data */
	tn_img.xmax = xsize; tn_img.ymax = ysize;
	tn_img.pixbuf = gdk_pixbuf_new_from_data ((unsigned char*)temp, GDK_COLORSPACE_RGB, FALSE, 8, 
														xsize, ysize, 3*xsize 		/* rowstride */,
														pixbuf_destroy_handler		/* destroy callback */,
														tn_img.pixbuf 				/* destroy callback user_data */);
	tn_img.width = xsize; tn_img.height = ysize;
	return 0;

}



void destroy_thumbnail ()		/* Don't really destroy. Rather merely hide it. */
{
	have_thumbnail = 0;

	if (tn_img.pixbuf) g_object_unref (G_OBJECT(tn_img.pixbuf));
	tn_img.pixbuf = NULL;

	gtk_widget_hide ( GTK_WIDGET ( tn_img.parent ) );

}




gboolean on_thumbnail_expose_event (GtkWidget *widget, GdkEventExpose  *event, gpointer user_data)
{
int x1,y1,x2,y2,w,h;

//	if(event->count > 0) 			/* Draw only last expose */
//	{
//		return TRUE;
//	}

	/* Clamp w,h back to image region - if anything unusual occurs, use full area */

	x1 = event->area.x; if (x1<0) x1=0; else if (x1>tn_img.width) x1=0;
	y1 = event->area.y; if (y1<0) y1=0; else if (y1>tn_img.height) y1=0;
	x2 = x1+event->area.width; y2 = y1+event->area.height;		/* Inclusive last row/column */
	if (x2 > tn_img.width) x2=tn_img.width;
	if (y2 > tn_img.height) y2=tn_img.height;
	w = x2-x1; h = y2-y1;
	dp (2, "Thumbnail expose. Area is %d, %d, %d, %d\n", x1,y1, w,h);

	gdk_pixbuf_render_to_drawable (tn_img.pixbuf, tn_img.drawable->window, 
									tn_img.drawable->style->fg_gc[GTK_STATE_NORMAL], 
									x1, y1,				/* src_x and src_y */
									x1, y1,				/* dest_x and dest_y */
									w, h,				/* width & height */
									GDK_RGB_DITHER_NORMAL, 0, 0);

  	return FALSE;
}




gboolean on_thumbnail_configure_event (GtkWidget *widget, GdkEventButton *event, gpointer  user_data)
{
int x,y;

	thumbnail_get_image_from_data (&mainimg);

	gdk_flush ();

	return FALSE;
}



/* OBSOLETE function to create a separate thumbnail window */

void create_thumbnail (void)
{
GtkWidget *tn_window;
GtkWidget *thumbnail;

	tn_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (tn_window), "tn_window", tn_window);
	gtk_window_set_title (GTK_WINDOW (tn_window), "Thumbnail");
	
	thumbnail = gtk_drawing_area_new ();
	gtk_widget_ref (thumbnail);
	gtk_object_set_data_full (GTK_OBJECT (tn_window), "thumbnail", thumbnail,
								(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (thumbnail);
	gtk_container_add (GTK_CONTAINER (tn_window), thumbnail);
	gtk_widget_set_events (thumbnail, GDK_BUTTON_PRESS_MASK);

	gtk_signal_connect (GTK_OBJECT (tn_window), "destroy",
						GTK_SIGNAL_FUNC (destroy_thumbnail),NULL);
//	gtk_signal_connect (GTK_OBJECT (thumbnail), "button_press_event",
//						GTK_SIGNAL_FUNC (on_thumbnail_button_press_event),NULL);
						
	gtk_signal_connect (GTK_OBJECT (thumbnail), "configure_event",
						GTK_SIGNAL_FUNC (on_thumbnail_configure_event),NULL);
	gtk_signal_connect (GTK_OBJECT (thumbnail), "expose_event",
						GTK_SIGNAL_FUNC (on_thumbnail_expose_event),NULL);
	
	gtk_widget_set_size_request (tn_window, TN_SIZE+10, TN_SIZE+10);
	gtk_window_set_resizable (GTK_WINDOW(tn_window), FALSE);		/* No user resizing allowed */
	gtk_widget_show (tn_window);
	have_thumbnail = 1; 

	gdk_window_set_cursor (thumbnail->window, crosshair_cursor);
	tn_img.parent = tn_window;
	tn_img.drawable = thumbnail;
	have_thumbnail = 1;

}




void display_thumbnail ()
{

	if (!file_ok || !mainimg.data) return;

	if (have_thumbnail) 
		if (tn_img.pixbuf) g_object_unref (G_OBJECT(tn_img.pixbuf));

	thumbnail_get_image_from_data (&mainimg);
	have_thumbnail = 1;
	gtk_widget_show (tn_img.parent);

	gtk_widget_queue_draw_area (tn_img.drawable, 0, 0, tn_img.width, tn_img.height);

}



