
/*
 *  secondwnd: Functions that use a secondary display, i.e. MIP and cut through a 3D stack
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "defs.h"
#include "imgbuf.h"
#include "callbacks.h"
#include "application.h"
#include "dlgbox.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "secondwnd.h"


/* TODO: The secondary image contains too many global variables that are
	self-contained in the s_img structure. */


GtkWidget *secondaryimg;
GtkWidget *secondary_status;
GtkWidget *secondary_label;


/******************************************************+

	EVENT CALLBACKS
	
********************************************************/


void close_secondary_image ()
{

	if (secondaryimage_exists)
	{
		freebuf (&s_img);
		if (s_img.pixbuf) g_object_unref (G_OBJECT(s_img.pixbuf));		/* If reference count drops tpo zero, pixbuf will be destroyed */

		secondaryimage_exists = 0;
		gtk_widget_destroy ( GTK_WIDGET ( secondwnd ) );
	}

}


void secondary_image_redraw ()
{
int w,h;


		w = s_img.width;
		h = s_img.height;
		dp (2, "Secondary image redraw: Scheduled expose event\n");
		gtk_widget_queue_draw_area (s_img.drawable, 0, 0, w, h);
}


void secondarypixelpeek (int x, int y); 	/* needs to be known in advance */


void second_statusbar_display (const char* text)
{

guint context;

	context = gtk_statusbar_get_context_id((GtkStatusbar*)secondary_status, "secondarystatus");

	if (sec_statusbar_was_used) /* pop the last message */
	{
		gtk_statusbar_pop((GtkStatusbar*)secondary_status, context);
		sec_statusbar_was_used=0;
	}
	
	gtk_statusbar_push((GtkStatusbar*)secondary_status, context, text);
	sec_statusbar_was_used = 1;
}


gboolean on_secondaryimg_expose_event (GtkWidget *widget, GdkEventExpose  *event, gpointer user_data)
{
int x,y;		/* To save some typing */
int w,h;

	x = event->area.x; y = event->area.y;
	w = event->area.width; h = event->area.height;

	/* If anything is out of range, force it back. This can happen if the display
		window is dragged to a size larger than the pixbuf area */

	if (x>s_img.width)
	{
		x = 0; w = s_img.width;
	}
	if (y>mainimg.height)
	{
		y = 0; h = s_img.height;
	}

	if (w>s_img.width) w=s_img.width;
	if (h>s_img.height) h=s_img.height;
	if (x+w>s_img.width) w=s_img.width-x;
	if (y+h>s_img.height) h=s_img.height-y;

	dp (2, "Secondary window expose event. Area is %d, %d, %d, %d\n", x, y, event->area.width, event->area.height);

	gdk_pixbuf_render_to_drawable(s_img.pixbuf, secondaryimg->window, secondaryimg->style->fg_gc[GTK_STATE_NORMAL], 
									x,y,					/* src_x and src_y */
									x,y, 					/* dest_x and dest_y */
									w, h,					/* width & height */
									GDK_RGB_DITHER_NORMAL, 0, 0);


  return FALSE;
}


gboolean on_secondaryimg_configure_event (GtkWidget *widget, GdkEventConfigure *event, gpointer user_data)
{

	gtk_widget_set_size_request (secondaryimg, s_img.width,s_img.height);
	dp (2, "Secondary window configure event\n");

	return FALSE;
}


gboolean on_secondaryimg_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
int x,y;
GdkModifierType state;

	gdk_window_get_pointer (widget->window, &x, &y, &state);
	secondarypixelpeek (x,y);


  	return FALSE;
}


gboolean on_secondaryimg_button_release_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{

  return FALSE;
}


void destroy_secondaryimg (GtkWidget *widget, gpointer user_data)
{
	close_secondary_image ();
}


/* TODO: Use the main save function, not the rgb-restricted imlib function */

void on_secondaryimg_save_clicked (GtkButton* savebutton, gpointer user_data)
{
char filename[1024];
gchar* selectedfile;

	selectedfile = ChooseFile("Save secondary image as (RGB only):");
	if (selectedfile)
	{
		strcpy (filename, (char*)selectedfile);
		if (save_file (filename, &s_img))
			second_statusbar_display ("Save operation failed");
		else
			second_statusbar_display ("Image saved.");
	}

}

void create_secondwnd (image_type *img, gchar* title)
{

GtkWidget *vbox1, *table;
GtkWidget* hbox1;
GtkWidget* swbutton1;
GtkWidget* swbutton2;
GtkWidget* hseparator1;
GtkWidget *scrolledwindow2;
GtkWidget *viewport2;


	secondwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (secondwnd), "secondwnd", secondwnd);
	gtk_window_set_title (GTK_WINDOW (secondwnd), "Secondary Display");
	gtk_window_set_default_size (GTK_WINDOW(secondwnd), 400,300);

	/* Items will be placed top to bottom into this vbox as they come */

	vbox1 = vbox_in_window (secondwnd, "svbox1");

	/* A label field with a title or caption */

	secondary_label = MakeLabel (secondwnd,vbox1,title, "secondary_label",-1,-1);

	/* the actual drawing area */

	scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (scrolledwindow2);
	gtk_container_add (GTK_CONTAINER (vbox1), scrolledwindow2);
	
	viewport2 = gtk_viewport_new (NULL, NULL);
	gtk_widget_show (viewport2);
	gtk_container_add (GTK_CONTAINER (scrolledwindow2), viewport2);

	table = gtk_table_new (2, 2, FALSE); 
	gtk_widget_ref (table);
	gtk_object_set_data_full (GTK_OBJECT (secondwnd), "sdspytbl", table, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (table);
	gtk_container_add (GTK_CONTAINER (viewport2), table);

	secondaryimg = gtk_drawing_area_new ();
	gtk_table_attach (GTK_TABLE (table), secondaryimg, 0, 1, 0, 1,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 
					4, 4);						/* H- and V- padding */
	gtk_object_set_data_full (GTK_OBJECT (secondwnd), "secondaryimg", secondaryimg,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (secondaryimg);

	/* a nice-looking hseparator */

	hseparator1 = MakeHsep (secondwnd,vbox1, "secondary_label",-1,-1);

	/* a hbox to contain the buttons */

	hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_widget_ref (hbox1);
	gtk_object_set_data_full (GTK_OBJECT (secondwnd), "swhbox1", hbox1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox1);
	/* gtk_container_add (GTK_CONTAINER (vbox1), hbox1); */
	gtk_box_pack_start (GTK_BOX (vbox1), hbox1, FALSE, FALSE, 2);
 
 	swbutton1 = Button_in_table (secondwnd,hbox1,"Save image","swbutton1",-1,-1,on_secondaryimg_save_clicked);
 	swbutton2 = Button_in_table (secondwnd,hbox1,"Dismiss","swbutton2",-1,-1,destroy_secondaryimg);
 
	/* And finally a status bar */

	secondary_status = gtk_statusbar_new ();
	gtk_widget_ref (secondary_status);
	gtk_object_set_data_full (GTK_OBJECT (secondwnd), "secondary_status", secondary_status,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (secondary_status);
	gtk_box_pack_start (GTK_BOX (vbox1), secondary_status, FALSE, FALSE, 0);

	gtk_widget_set_events (secondaryimg, GDK_EXPOSURE_MASK
										  | GDK_LEAVE_NOTIFY_MASK
										  | GDK_BUTTON_PRESS_MASK
										  | GDK_BUTTON_RELEASE_MASK
										  | GDK_POINTER_MOTION_MASK
										  | GDK_POINTER_MOTION_HINT_MASK
										  | GDK_KEY_PRESS_MASK);


	gtk_signal_connect (GTK_OBJECT (secondaryimg), "expose_event",
                    	GTK_SIGNAL_FUNC (on_secondaryimg_expose_event),NULL);
	gtk_signal_connect (GTK_OBJECT (secondaryimg), "button_press_event",
                    	GTK_SIGNAL_FUNC (on_secondaryimg_button_press_event),NULL);
	gtk_signal_connect (GTK_OBJECT (secondaryimg), "button_release_event",
                    	GTK_SIGNAL_FUNC (on_secondaryimg_button_release_event),NULL);
	gtk_signal_connect (GTK_OBJECT (secondaryimg), "configure_event",
                    	GTK_SIGNAL_FUNC (on_secondaryimg_configure_event),NULL);
	gtk_signal_connect (GTK_OBJECT (secondaryimg), "destroy",
                    	GTK_SIGNAL_FUNC (destroy_secondaryimg),secondwnd);


	sec_statusbar_was_used=0;
	gtk_widget_show (secondwnd);
	img->width = img->xmax;
	img->height = img->ymax;			/* Drawing area allocation size */
	img->parent = secondwnd;			/* The secondary image's parent window */
	img->drawable = secondaryimg;		/* The drawingarea */

	gtk_widget_set_size_request (img->drawable, img->width, img->height);
	gdk_window_set_cursor (secondaryimg->window, crosshair_cursor);

}


/* This is an exact copy of the pixelpeek of the primary image */

void secondarypixelpeek (int x, int y)
{
char s[255];
int ix,iy,iz;
long zz;

	if (!s_img.data) return;	/* No image, no peeking */


	ix = secondaryimg->allocation.width;
	iy = secondaryimg->allocation.height;
	iz = 0;	/* secondary image is always 2D */

	ix = IROUND((double)(x) * (double)s_img.xmax/(double)ix);
	iy = IROUND((double)(y) * (double)s_img.ymax/(double)iy);

	dp (1, "Click at %d,%d rescaled through %d,%d to %d,%d\n",x,y,secondaryimg->allocation.width,secondaryimg->allocation.height,ix,iy);

	zz = ix + s_img.xmax*(iy + s_img.ymax*iz);
	switch (s_img.imgtype)
	{
		case B_CHAR: 
		{
			unsigned char* p = s_img.data;
			sprintf (s, " (%d,%d) Data=%d", ix,iy,p[zz]);
			break;
		}
		case B_SHORT: 
		{
			short* p = s_img.data;
			sprintf (s, " (%d,%d) Data=%d", ix,iy,p[zz]);
			break;
		}
		case B_RGB: 
		{
			unsigned char* p = s_img.data;
			zz *= 3;	/* because we have three chars per pixel */
			sprintf (s, " (%d,%d) R=%d G=%d B=%d",ix,iy,p[zz],p[zz+1],p[zz+2]);
			break;
		}
		case B_FLOAT: 
		{
			float* p = s_img.data;
			sprintf (s, " (%d,%d) Data=%10.4f", ix,iy,p[zz]);
			break;
		}
	}
	

	second_statusbar_display (s);
	
}




/* Use the secondary display to display any image passed here */

void secondary_display (image_type* img, int zslice, char* title)
{
long e, i, idx, zofs;
rgbtriplet*	tempimg;


	/* free existing image areas to prevent mem leaking 
	  -- double-check in a quiet moment - may cause trouble */
	
	if (s_img.data) freebuf (&s_img);
	if (s_img.pixbuf) g_object_unref (GTK_OBJECT(s_img.pixbuf));
	
	idx = img->xmax * img->ymax; zofs = idx * zslice;
	
	/* Clone current slice from the image passed to this function. We do this to save the
		underlying data for secondary-window pixel operations */

	if (allocate_image (&s_img, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
	e = DATALEN(img->imgtype);
	for (i=0; i < e*idx; i++)
		((char*)s_img.data)[i] = ((char*)img->data)[i+zofs*e];		/* Clones a slice */
	make_minmax (&s_img);
	
	/* Prepare an RGB buffer to create a pixbuf from */

	tempimg = (rgbtriplet*) calloc (idx, sizeof (rgbtriplet));
	if (!tempimg) fatal_memerr ();

	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p;
			p = img->data;
			for (i=0; i<idx; i++)
			{
				tempimg[i].red = p[i+zofs];
				tempimg[i].green = p[i+zofs];
				tempimg[i].blue = p[i+zofs];
			}
			break;
		}
		case B_SHORT:
		{
			short* p;
			p = img->data;
			for (i=0; i<idx; i++)
			{
				tempimg[i].red = tempimg[i].green = tempimg[i].blue = 
					IROUND(255.0*(double)(p[i+zofs] - img->minval.imin)
							/ (double) (img->maxval.imax-img->minval.imin));
			}
			break;
		}
		case B_RGB:
		{
			rgbtriplet* p;
			p = img->data;
			for (i=0; i<idx; i++)
				tempimg[i] = p[i+zofs];
			break;
		}
		case B_FLOAT:
		{
			float* p;
			p = img->data;
			for (i=0; i<idx; i++)
			{
				tempimg[i].red = tempimg[i].green = tempimg[i].blue = 
					IROUND(255.0*(p[i+zofs] - img->minval.fmin)
							/ (img->maxval.fmax-img->minval.fmin));
			}
			break;
		}
	}


	/* Now create a pixbuf from the temp data */
	
	s_img.pixbuf = gdk_pixbuf_new_from_data ((unsigned char*)tempimg, GDK_COLORSPACE_RGB, FALSE, 8, 
														s_img.xmax, s_img.ymax, 3*s_img.xmax /* rowstride */,
														pixbuf_destroy_handler /* destroy callback */,
														s_img.pixbuf /* destroy callback user_data */);

	
	/* Display the newly created image */
				

	if (!secondaryimage_exists) 
		create_secondwnd (&s_img, title);
	else
		gtk_label_set_text (GTK_LABEL(secondary_label), title);

	secondaryimage_exists = 1;
	secondary_image_redraw ();
}


